""" Generator functions related to trial and test functions and the accumulation loop"""

from dune.perftool.generation import domain, iname, pymbolic_expr, symbol, globalarg
from dune.perftool.ufl.modified_terminals import ModifiedArgumentDescriptor
from dune.perftool.pdelab import name_index
from dune.perftool.pdelab.basis import (evaluate_trialfunction,
                                        evaluate_trialfunction_gradient,
                                        lfs_iname,
                                        name_basis,
                                        name_basis_gradient,
                                        name_lfs_bound,
                                        )

from pymbolic.primitives import Subscript, Variable


@symbol
def name_testfunction_gradient(ma):
    assert ma.grad

    return name_basis_gradient(ma.argexpr.element())


@symbol
def name_testfunction(ma):
    assert not ma.grad

    return name_basis(ma.argexpr.element())


@pymbolic_expr
def pymbolic_testfunction(ma):
    # we only accept an index if we treat a gradient...
    assert bool(ma.index) == ma.grad

    if ma.grad:
        return Subscript(Variable(name_testfunction_gradient(ma)), (Variable(lfs_iname(ma.argexpr.element())), Variable(name_index(ma.index))))
    else:
        return Subscript(Variable(name_testfunction(ma)), Variable(lfs_iname(ma.argexpr.element())))


@symbol
def name_trialfunction_gradient(ma):
    assert ma.grad

    evaluate_trialfunction_gradient(ma.argexpr.element(), "gradu")
    return "gradu"


@symbol
def name_trialfunction(ma):
    assert not ma.grad

    # TODO trigger the evaluation of the trial function
    evaluate_trialfunction(ma.argexpr.element(), "u")
    return "u"


@pymbolic_expr
def pymbolic_trialfunction(ma):
    # we only accept an index if we treat a gradient...
    assert bool(ma.index) == ma.grad

    if ma.grad:
        return Subscript(Variable(name_trialfunction_gradient(ma)), Variable(name_index(ma.index)))
    else:
        return Variable(name_trialfunction(ma))


@symbol
def name_testfunctionspace():
    return "lfsv"


@symbol
def name_trialfunctionspace():
    return "lfsu"


@symbol
def type_testfunctionspace():
    return "LFSV"


@symbol
def type_trialfunctionspace():
    return "LFSU"


@symbol
def name_coefficientcontainer():
    return "x"


@symbol
def type_coefficientcontainer():
    return "X"


def name_argumentspace(ma):
    if ma.argexpr.number() == 0:
        return name_testfunctionspace(ma)
    if ma.argexpr.number() == 1:
        return name_trialfunctionspace(ma)
    # We should never encounter an argument other than 0 or 1
    assert False


def name_argument(ma):
    if ma.argexpr.number() == 0:
        return name_testfunction(ma)
    if ma.argexpr.number() == 1:
        return name_trialfunction(ma)
    # We should never encounter an argument other than 0 or 1
    assert False


def pymbolic_argument(ma):
    if ma.argexpr.number() == 0:
        return pymbolic_testfunction(ma)
    if ma.argexpr.number() == 1:
        return pymbolic_trialfunction(ma)
    assert False


@symbol
def name_jacobian():
    return "jac"


@symbol
def type_jacobian():
    return "J"


@symbol
def name_residual():
    return "r"


@symbol
def type_residual():
    return "R"


def name_accumulation_variable():
    from dune.perftool.generation import get_global_context_value
    ft = get_global_context_value("form_type")
    if ft == 'residual':
        return name_residual()
    if ft == 'jacobian':
        return name_jacobian()
    assert False


def type_accumulation_variable():
    from dune.perftool.generation import get_global_context_value
    ft = get_global_context_value("form_type")
    if ft == 'residual':
        return type_residual()
    if ft == 'jacobian':
        return type_jacobian()
    assert False
